@@ -1,6 +1,6 @@
 /* 
    Higher Level Interface to XML Parsers.
-   Copyright (C) 1999-2001, Joe Orton <joe@light.plus.com>
+   Copyright (C) 1999-2002, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -36,34 +36,41 @@
 #include "ne_utils.h"
 #include "ne_string.h"
 
-#ifdef HAVE_EXPAT
+#if defined(HAVE_EXPAT)
 
-/* Pick up the right header. */
 #if defined(HAVE_EXPAT_H)
 #include <expat.h>
 #elif defined(HAVE_XMLPARSE_H)
 #include <xmlparse.h>
 #elif defined(HAVE_XMLTOK_XMLPARSE_H)
 #include <xmltok/xmlparse.h>
+#else
+#error no expat header known!
 #endif
 
 typedef XML_Char ne_xml_char;
 
-#else /* not HAVE_EXPAT */
-# ifdef HAVE_LIBXML
+#elif defined(HAVE_LIBXML)
 
-#ifdef HAVE_XMLVERSION_H
-/* libxml2: pick up the version string. */
+/* pick up the version string. */
+#if defined(HAVE_XMLVERSION_H)
 #include <xmlversion.h>
+#elif defined(HAVE_LIBXML_XMLVERSION_H)
+#include <libxml/xmlversion.h>
+#endif
+
+#if defined(HAVE_PARSER_H)
+# include <parser.h>
+#elif defined(HAVE_LIBXML_PARSER_H)
+# include <libxml/parser.h>
+#else
+#error no libxml header known!
 #endif
 
-/******** libxml **********/
-#  include <parser.h>
 typedef xmlChar ne_xml_char;
 
-# else /* not HAVE_LIBXML */
+#else /* not HAVE_LIBXML */
 #  error need an XML parser
-# endif /* not HAVE_LIBXML */
 #endif /* not HAVE_EXPAT */
 
 /* Approx. one screen of text: */
@@ -103,12 +110,6 @@ struct ne_xml_state {
     struct ne_xml_state *parent; /* The parent in the tree */
 };
 
-  
-/* TODO: 
- * could move 'valid' into state, maybe allow optional
- * continuation past an invalid branch.
- */
-
 /* We pass around a ne_xml_parser as the userdata in the parsing
  * library.  This maintains the current state of the parse and various
  * other bits and bobs. Within the parse, we store the current branch
@@ -149,7 +150,7 @@ static const char *friendly_name(const struct ne_xml_elm *elm)
     }
 }
 
-const static struct ne_xml_elm root_element = 
+static const struct ne_xml_elm root_element = 
 { "@<root>@", NE_ELM_root, 0 };
 
 /* The callback handlers */
@@ -157,7 +158,11 @@ static void start_element(void *userdata, const ne_xml_char *name, const ne_xml_
 static void end_element(void *userdata, const ne_xml_char *name);
 static void char_data(void *userdata, const ne_xml_char *cdata, int len);
 
+/* define this for expat, libxml 2.x. libxml 1.x is not good at UTF-8
+ * stuff. */
+#if !defined (HAVE_LIBXML) || defined(LIBXML_VERSION_STRING)
 #define NE_XML_DECODE_UTF8
+#endif
 
 #ifdef NE_XML_DECODE_UTF8
 
@@ -269,9 +274,9 @@ static int find_handler(ne_xml_parser *p, struct ne_xml_state *state)
     int n, got_unknown = 0;
 
     for (cur = state->parent->handler; cur != NULL; cur = cur->next) {
-	for (n = 0; (cur->elements[n].nspace != NULL || (
-		     cur->elements[n].nspace == NULL && 
-		     cur->elements[n].id == NE_ELM_unknown)); n++) {
+	for (n = 0; (cur->elements[n].nspace != NULL ||
+		     (cur->elements[n].nspace == NULL && 
+		      cur->elements[n].id == NE_ELM_unknown)); n++) {
 	    if (cur->elements[n].nspace != NULL && 
 		(strcasecmp(cur->elements[n].name, name) == 0 && 
 		 strcasecmp(cur->elements[n].nspace, nspace) == 0)) {
@@ -286,10 +291,10 @@ static int find_handler(ne_xml_parser *p, struct ne_xml_state *state)
 		    return 0;
 		case NE_XML_INVALID:
 		    NE_DEBUG(NE_DBG_XML, "Invalid context.\n");
-		    snprintf(p->error, ERR_SIZE, 
-			     _("XML is not valid (%s found in parent %s)"),
-			     friendly_name(&cur->elements[n]), 
-			     friendly_name(state->parent->elm));
+		    ne_snprintf(p->error, ERR_SIZE, 
+				_("XML is not valid (%s found in parent %s)"),
+				friendly_name(&cur->elements[n]), 
+				friendly_name(state->parent->elm));
 		    return -1;
 		default:
 		    /* ignore it */
@@ -309,10 +314,10 @@ static int find_handler(ne_xml_parser *p, struct ne_xml_state *state)
 		    break;
 		case NE_XML_INVALID:
 		    NE_DEBUG(NE_DBG_XML, "Invalid context.\n");
-		    snprintf(p->error, ERR_SIZE, 
-			     _("XML is not valid (%s found in parent %s)"),
-			     friendly_name(&cur->elements[n]), 
-			     friendly_name(state->parent->elm));
+		    ne_snprintf(p->error, ERR_SIZE, 
+				_("XML is not valid (%s found in parent %s)"),
+				friendly_name(&cur->elements[n]), 
+				friendly_name(state->parent->elm));
 		    return -1;
 		default:
 		    NE_DEBUG(NE_DBG_XML, "Declined by handler.\n");
@@ -329,22 +334,23 @@ static int find_handler(ne_xml_parser *p, struct ne_xml_state *state)
 	return 0;
     } else {
 	NE_DEBUG(NE_DBG_XMLPARSE, "Unexpected element\n");
-	snprintf(p->error, ERR_SIZE, 
-		 _("Unknown XML element `%s (in %s)'"), name, nspace);
+	ne_snprintf(p->error, ERR_SIZE, 
+		    _("Unknown XML element `%s (in %s)'"), name, nspace);
 	return -1;
     }
 }
 
 /* Called with the start of a new element. */
-static void 
-start_element(void *userdata, const ne_xml_char *name, const ne_xml_char **atts) 
+static void start_element(void *userdata, const ne_xml_char *name,
+			  const ne_xml_char **atts) 
 {
     ne_xml_parser *p = userdata;
     struct ne_xml_state *s;
 
     if (!p->valid) {
 	/* We've stopped parsing */
-	NE_DEBUG(NE_DBG_XML, "Parse died. Ignoring start of element: %s\n", name);
+	NE_DEBUG(NE_DBG_XML, "Parse died. Ignoring start of element: %s\n",
+		 name);
 	return;
     }
 
@@ -361,8 +367,8 @@ start_element(void *userdata, const ne_xml_char *name, const ne_xml_char **atts)
 	if (atts != NULL) {
 	    int n;
 	    for (n = 0; atts[n] != NULL; n+=2) {
-		ne_buffer_concat(p->buffer, " ", atts[n], "=\"", atts[n+1], "\"",
-			       NULL);
+		ne_buffer_concat(p->buffer, " ", atts[n], 
+				 "=\"", atts[n+1], "\"", NULL);
 	    }
 	}
 	ne_buffer_zappend(p->buffer, ">");
@@ -385,8 +391,8 @@ start_element(void *userdata, const ne_xml_char *name, const ne_xml_char **atts)
 
     /* Map the element name to an id */
     NE_DEBUG(NE_DBG_XML, "Mapping element name {%s}%s... ", 
-	  s->elm_real.nspace, s->elm_real.name);
-
+	     s->elm_real.nspace, s->elm_real.name);
+    
     if (find_handler(p, s)) {
 	p->valid = 0;
 	return;
@@ -401,9 +407,14 @@ start_element(void *userdata, const ne_xml_char *name, const ne_xml_char **atts)
     /* Is this element using mixed-mode? */
     s->mixed = ((s->elm->flags & NE_XML_MIXED) == NE_XML_MIXED);
 
+    if (!s->mixed) {
+	/* clear cdata buffer. */
+	ne_buffer_clear(p->buffer);
+    }
+
     if (s->handler->startelm_cb) {
-	if ((*s->handler->startelm_cb)(s->handler->userdata, s->elm, 
-				       (const char **) atts)) {
+	if (s->handler->startelm_cb(s->handler->userdata, s->elm, 
+				    (const char **) atts)) {
 	    NE_DEBUG(NE_DBG_XML, "Startelm callback failed.\n");
 	    p->valid = 0;
 	}
@@ -417,7 +428,6 @@ start_element(void *userdata, const ne_xml_char *name, const ne_xml_char **atts)
 static void destroy_state(struct ne_xml_state *s) 
 {
     struct ne_xml_nspace *this_ns, *next_ns;
-    NE_DEBUG(NE_DBG_XMLPARSE, "Freeing namespaces...\n");
     NE_FREE(s->default_ns);
     NE_FREE(s->real_name);
     /* Free the namespaces */
@@ -429,7 +439,6 @@ static void destroy_state(struct ne_xml_state *s)
 	free(this_ns);
 	this_ns = next_ns;
     };
-    NE_DEBUG(NE_DBG_XMLPARSE, "Finished freeing namespaces.\n");
     free(s);
 }
 
@@ -438,8 +447,8 @@ static void char_data(void *userdata, const ne_xml_char *data, int len)
     ne_xml_parser *p = userdata;
     
     if (p->current->mixed) {
-	(*p->current->handler->cdata_cb)( 
-	    p->current->handler->userdata, p->current->elm, data, len);
+	p->current->handler->cdata_cb(p->current->handler->userdata,
+				      p->current->elm, data, len);
 	return;
     }
 
@@ -447,7 +456,8 @@ static void char_data(void *userdata, const ne_xml_char *data, int len)
     /* First, if this is the beginning of the CDATA, skip all
      * leading whitespace, we don't want it. */
     NE_DEBUG(NE_DBG_XMLPARSE, "Given %d bytes of cdata.\n", len);
-    if (ne_buffer_size(p->buffer) == 0) {
+    if ((p->current->elm->flags & NE_XML_STRIPWS) &&
+	ne_buffer_size(p->buffer) == 0) {
 	int wslen = 0;
 	/* Ignore any leading whitespace */
 	while (wslen < len && 
@@ -458,7 +468,7 @@ static void char_data(void *userdata, const ne_xml_char *data, int len)
 	data += wslen;
 	len -= wslen;
 	NE_DEBUG(NE_DBG_XMLPARSE, "Skipped %d bytes of leading whitespace.\n", 
-	       wslen);
+		 wslen);
 	if (len == 0) {
 	    NE_DEBUG(NE_DBG_XMLPARSE, "Zero bytes of content.\n");
 	    return;
@@ -523,7 +533,8 @@ static void end_element(void *userdata, const ne_xml_char *name)
     struct ne_xml_state *s = p->current;
     if (!p->valid) {
 	/* We've stopped parsing */
-	NE_DEBUG(NE_DBG_XML, "Parse died. Ignoring end of element: %s\n", name);
+	NE_DEBUG(NE_DBG_XML, "Parse died. Ignoring end of element: %s\n", 
+		 name);
 	return;
     }
     if (p->collect > 0) {
@@ -541,22 +552,46 @@ static void end_element(void *userdata, const ne_xml_char *name)
 	
     /* process it */
     if (s->handler->endelm_cb) {
-	NE_DEBUG(NE_DBG_XMLPARSE, "Calling endelm callback for %s.\n", s->elm->name);
-	if ((*s->handler->endelm_cb)(s->handler->userdata, s->elm,
-				     p->want_cdata?p->buffer->data:NULL)) {
+	NE_DEBUG(NE_DBG_XMLPARSE, "Calling endelm callback for %s.\n",
+		 s->elm->name);
+	if (s->handler->endelm_cb(s->handler->userdata, s->elm,
+				  p->want_cdata?p->buffer->data:NULL)) {
 	    NE_DEBUG(NE_DBG_XML, "Endelm callback failed.\n");
 	    p->valid = 0;
 	}
     }
     p->current = s->parent;
     /* Move the current pointer up the branch */
-    NE_DEBUG(NE_DBG_XML, "Back in element: %s\n", friendly_name(p->current->elm));
+    NE_DEBUG(NE_DBG_XML, "Back in element: %s\n", 
+	     friendly_name(p->current->elm));
     if (p->want_cdata) {
 	ne_buffer_clear(p->buffer);
     } 
     destroy_state(s);
 }
 
+/* Find a namespace definition for 'prefix' in given parser state,
+ * where length of prefix is 'pfxlen'.  Returns the URI or NULL. */
+static const char *
+resolve_nspace(const struct ne_xml_state *state, 
+	       const char *prefix, size_t pfxlen)
+{
+    const struct ne_xml_state *s;
+
+    /* Search up the tree. */
+    for (s = state; s != NULL; s = s->parent) {
+	const struct ne_xml_nspace *ns;
+	/* Iterate over defined spaces on this node. */
+	for (ns = s->nspaces; ns != NULL; ns = ns->next) {
+	    if (strlen(ns->name) == pfxlen && 
+		memcmp(ns->name, prefix, pfxlen) == 0)
+		return ns->uri;
+	}
+    }
+	
+    return NULL;
+}
+
 /* Parses the attributes, and handles XML namespaces. 
  * With a little bit of luck.
  * Returns:
@@ -576,13 +611,27 @@ static int parse_element(ne_xml_parser *p, struct ne_xml_state *state,
     if (atts != NULL) {
 	int attn;
 	for (attn = 0; atts[attn]!=NULL; attn+=2) {
-	    NE_DEBUG(NE_DBG_XMLPARSE, "Got attribute: [%s] = [%s]\n", atts[attn], atts[attn+1]);
+	    NE_DEBUG(NE_DBG_XMLPARSE, "Got attribute: [%s] = [%s]\n", 
+		     atts[attn], atts[attn+1]);
 	    if (strcasecmp(atts[attn], "xmlns") == 0) {
 		/* New default namespace */
 		state->default_ns = ne_strdup(atts[attn+1]);
 		NE_DEBUG(NE_DBG_XMLPARSE, "New default namespace: %s\n", 
-		       state->default_ns);
+			 state->default_ns);
+		/* FIXME: if attribute value is empty, then this means
+		 * "no more default namespace". 
+		 * http://www.w3.org/TR/1999/REC-xml-names-19990114/#ns-decl
+		 */
 	    } else if (strncasecmp(atts[attn], "xmlns:", 6) == 0) {
+		if (atts[attn+1][0] == '\0') {
+		    NE_DEBUG(NE_DBG_XMLPARSE,
+			     "Invalid namespace declaration.\n");
+		    ne_snprintf(p->error, ERR_SIZE, 
+				"Invalid namespace declaration for '%s' "
+				"in '%s' at line %d.",
+				atts[attn], name, ne_xml_currentline(p));
+		    return -1;
+		}
 		/* New namespace scope */
 		ns = ne_calloc(sizeof(struct ne_xml_nspace));
 		ns->next = state->nspaces;
@@ -590,7 +639,7 @@ static int parse_element(ne_xml_parser *p, struct ne_xml_state *state,
 		ns->name = ne_strdup(atts[attn]+6); /* skip the xmlns= */
 		ns->uri = ne_strdup(atts[attn+1]);
 		NE_DEBUG(NE_DBG_XMLPARSE, "New namespace scope: %s -> %s\n",
-		       ns->name, ns->uri);
+			 ns->name, ns->uri);
 	    }
 	}
     }
@@ -612,35 +661,26 @@ static int parse_element(ne_xml_parser *p, struct ne_xml_state *state,
 	}
     } else {
 	NE_DEBUG(NE_DBG_XMLPARSE, "Got namespace scope. Trying to resolve...");
-	/* Have a scope - resolve it */
-	for (xmlt = state; state->elm_real.nspace==NULL && xmlt!=NULL; xmlt=xmlt->parent) {
-	    for (ns = xmlt->nspaces; ns!=NULL; ns=ns->next) {
-		/* Just compare against the bit before the :
-		 * pnt points to the colon. */
-		if (strncasecmp(ns->name, name, pnt-name) == 0) {
-		    /* Scope matched! Hoorah */
-		    state->elm_real.nspace = ns->uri;
-		    /* end the search */
-		    break;
-		}
-	    }
-	}
+	state->elm_real.nspace = resolve_nspace(state, name, pnt-name);
+
 	if (state->elm_real.nspace != NULL) {
-	    NE_DEBUG(NE_DBG_XMLPARSE, "Resolved prefix to [%s]\n", state->elm_real.nspace);
+	    NE_DEBUG(NE_DBG_XMLPARSE, "Resolved prefix to [%s]\n",
+		     state->elm_real.nspace);
 	    /* The name is everything after the ':' */
 	    if (pnt[1] == '\0') {
-		snprintf(p->error, ERR_SIZE, 
-			  "Element name missing in '%s' at line %d.",
-			  name, ne_xml_currentline(p));
-		NE_DEBUG(NE_DBG_XMLPARSE, "No element name after ':'. Failed.\n");
+		ne_snprintf(p->error, ERR_SIZE, 
+			    "Element name missing in '%s' at line %d.",
+			    name, ne_xml_currentline(p));
+		NE_DEBUG(NE_DBG_XMLPARSE,
+			 "No element name after ':'. Failed.\n");
 		return -1;
 	    }
 	    state->real_name = ne_strdup(pnt+1);
 	} else {
 	    NE_DEBUG(NE_DBG_XMLPARSE, "Undeclared namespace.\n");
-	    snprintf(p->error, ERR_SIZE, 
-		      "Undeclared namespace in '%s' at line %d.",
-		      name, ne_xml_currentline(p));
+	    ne_snprintf(p->error, ERR_SIZE, 
+			"Undeclared namespace in '%s' at line %d.",
+			name, ne_xml_currentline(p));
 	    return -1;
 	}
     }
@@ -679,7 +719,6 @@ ne_xml_parser *ne_xml_create(void)
 static void push_handler(ne_xml_parser *p,
 			 struct ne_xml_handler *handler)
 {
-
     /* If this is the first handler registered, update the
      * base pointer too. */
     if (p->top_handlers == NULL) {
@@ -692,11 +731,11 @@ static void push_handler(ne_xml_parser *p,
 }
 
 void ne_xml_push_handler(ne_xml_parser *p,
-			  const struct ne_xml_elm *elements, 
-			  ne_xml_validate_cb validate_cb, 
-			  ne_xml_startelm_cb startelm_cb, 
-			  ne_xml_endelm_cb endelm_cb,
-			  void *userdata)
+			 const struct ne_xml_elm *elements, 
+			 ne_xml_validate_cb validate_cb, 
+			 ne_xml_startelm_cb startelm_cb, 
+			 ne_xml_endelm_cb endelm_cb,
+			 void *userdata)
 {
     struct ne_xml_handler *hand = ne_calloc(sizeof(struct ne_xml_handler));
 
@@ -744,7 +783,8 @@ void ne_xml_parse(ne_xml_parser *p, const char *block, size_t len)
     int ret, flag;
     /* duck out if it's broken */
     if (!p->valid) {
-	NE_DEBUG(NE_DBG_XML, "Not parsing %d bytes.\n", len);
+	NE_DEBUG(NE_DBG_XML, "Not parsing %" NE_FMT_SIZE_T " bytes.\n", 
+		 len);
 	return;
     }
     if (len == 0) {
@@ -752,7 +792,8 @@ void ne_xml_parse(ne_xml_parser *p, const char *block, size_t len)
 	block = "";
 	NE_DEBUG(NE_DBG_XML, "Got 0-length buffer, end of document.\n");
     } else {	
-	NE_DEBUG(NE_DBG_XML, "Parsing %d length buffer.\n", len);
+	NE_DEBUG(NE_DBG_XML, "Parsing %" NE_FMT_SIZE_T " length buffer.\n",
+		 len);
 	flag = 0;
     }
     /* Note, don't write a parser error if !p->valid, since an error
@@ -761,10 +802,10 @@ void ne_xml_parse(ne_xml_parser *p, const char *block, size_t len)
     ret = XML_Parse(p->parser, block, len, flag);
     NE_DEBUG(NE_DBG_XMLPARSE, "XML_Parse returned %d\n", ret);
     if (ret == 0 && p->valid) {
-	snprintf(p->error, ERR_SIZE,
-		  "XML parse error at line %d: %s", 
-		  XML_GetCurrentLineNumber(p->parser),
-		  XML_ErrorString(XML_GetErrorCode(p->parser)));
+	ne_snprintf(p->error, ERR_SIZE,
+		    "XML parse error at line %d: %s", 
+		    XML_GetCurrentLineNumber(p->parser),
+		    XML_ErrorString(XML_GetErrorCode(p->parser)));
 	p->valid = 0;
     }
 #else
@@ -772,8 +813,8 @@ void ne_xml_parse(ne_xml_parser *p, const char *block, size_t len)
     NE_DEBUG(NE_DBG_XMLPARSE, "xmlParseChunk returned %d\n", ret);
     if (p->parser->errNo && p->valid) {
 	/* FIXME: error handling */
-	snprintf(p->error, ERR_SIZE, "XML parse error at line %d.", 
-		 ne_xml_currentline(p));
+	ne_snprintf(p->error, ERR_SIZE, "XML parse error at line %d.", 
+		    ne_xml_currentline(p));
 	p->valid = 0;
     }
 #endif
@@ -816,7 +857,7 @@ void ne_xml_destroy(ne_xml_parser *p)
 
 void ne_xml_set_error(ne_xml_parser *p, const char *msg)
 {
-    snprintf(p->error, ERR_SIZE, msg);
+    ne_snprintf(p->error, ERR_SIZE, msg);
 }
 
 #ifdef HAVE_LIBXML
@@ -827,12 +868,12 @@ static void sax_error(void *ctx, const char *msg, ...)
     char buf[1024];
 
     va_start(ap, msg);
-    vsnprintf(buf, 1024, msg, ap);
+    ne_vsnprintf(buf, 1024, msg, ap);
     va_end(ap);
 
-    snprintf(p->error, ERR_SIZE, 
-	     _("XML parse error at line %d: %s."),
-	     p->parser->input->line, buf);
+    ne_snprintf(p->error, ERR_SIZE, 
+		_("XML parse error at line %d: %s."),
+		p->parser->input->line, buf);
 
     p->valid = 0;
 }
@@ -843,8 +884,9 @@ const char *ne_xml_get_error(ne_xml_parser *p)
     return p->error;
 }
 
-
-const char *ne_xml_get_attr(const char **attrs, const char *name)
+const char *
+ne_xml_get_attr(ne_xml_parser *p, const char **attrs, 
+		const char *nspace, const char *name)
 {
     int n;
     
@@ -858,8 +900,19 @@ const char *ne_xml_get_attr(const char **attrs, const char *name)
 #endif
 
     for (n = 0; attrs[n] != NULL; n += 2) {
-	if (strcmp(attrs[n], name) == 0) {
+	char *pnt = strchr(attrs[n], ':');
+
+	if (!nspace && !pnt && strcmp(attrs[n], name) == 0) {
 	    return attrs[n+1];
+	} else if (nspace && pnt) {
+	    /* If a namespace is given, and the local part matches,
+	     * then resolve the namespace and compare that too. */
+	    if (pnt && strcmp(pnt + 1, name) == 0) {
+		const char *uri = resolve_nspace(p->current, 
+						 attrs[n], pnt - attrs[n]);
+		if (uri && strcmp(uri, nspace) == 0)
+		    return attrs[n+1];
+	    }
 	}
     }
 
