@@ -64,7 +64,6 @@
 #include <ne_socket.h>
 #include <ne_locks.h>
 #include <ne_alloc.h>
-#include <ne_redirect.h>
 
 #include "common.h"
 
@@ -81,13 +80,15 @@ static netrc_entry *netrc_list; /* list of netrc entries */
 #endif
 
 ne_session *session;
-ne_lock_session *lock_session;
-static nssl_context *ssl_context;
+ne_lock_store *lock_store = NULL;
+const char *lock_store_fn = NULL;
 
-static char *progname, *lockstore; /* argv[0] */
+static char *progname; /* argv[0] */
 
-char *server_hostname, *proxy_hostname;
-int server_port, proxy_port;
+ne_uri server = {0}; /* current URI of server. */
+
+char *proxy_hostname;
+int proxy_port;
 char *server_username = NULL, *server_password = NULL;
 
 /* Global options */
@@ -116,11 +117,6 @@ static enum out_state {
 
 static RETSIGTYPE quit_handler(int signo);
 
-static void close_connection(void);
-static void open_connection(const char *url);
-
-static void execute_version(void);
-
 static void transfer_progress(void *ud, off_t progress, off_t total);
 static void connection_status(void *ud, ne_conn_status status, 
 			      const char *info);
@@ -129,98 +125,6 @@ static int supply_creds_server(void *userdata, const char *realm, int attempt,
 static int supply_creds_proxy(void *userdata, const char *realm, int attempt,
 			      char *username, char *password);
 
-#define CMD_VARY 9999
-
-/* Separate structures for commands and command names. */
-/* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND */
-const struct command commands[] = {
-/* C1: connected, 1-arg function C2: connected, 2-arg function
- * U0: disconnected, 0-arg function. */
-#define C1(x,c,h) { cmd_##x, true, 1, 1, parmscope_remote, execute_##x,c,h }
-#define U0(x,h) { cmd_##x, false, 0, 0, parmscope_none, execute_##x,#x,h }
-#define UO1(x,c,h) { cmd_##x, false, 0, 1, parmscope_none, execute_##x,c,h }
-#define C2M(x,c,h) { cmd_##x, true, 2, CMD_VARY, parmscope_remote, multi_##x,c,h }
-#define C1M(x,c,h) { cmd_##x, true, 1, CMD_VARY, parmscope_remote, multi_##x,c,h }
-    { cmd_ls, true, 0, 1, parmscope_remote, execute_ls, 
-      N_("ls [path]"), N_("List contents of current [or other] collection") },
-    C1(cd, N_("cd path"), N_("Change to specified collection")),
-    { cmd_pwd, true, 0, 0, parmscope_none, execute_pwd,
-      "pwd", N_("Display name of current collection") },
-    { cmd_put, true, 1, 2, parmscope_none, execute_put, 
-      N_("put local [remote]"), N_("Upload local file") },
-    { cmd_get, true, 1, 2, parmscope_none, execute_get, 
-      N_("get remote [local]"), N_("Download remote resource") },
-    C1M(mget, N_("mget remote..."), N_("Download many remote resources")),
-    { cmd_mput, true, 1, CMD_VARY, parmscope_local, multi_mput, 
-      N_("mput local..."), N_("Upload many local files") },
-    C1(edit, N_("edit resource"), N_("Edit given resource")),
-    C1M(less, N_("less remote..."), N_("Display remote resource through pager")), 
-    C1M(mkcol, N_("mkcol remote..."), N_("Create remote collection(s)")), 
-    C1M(cat, N_("cat remote..."), N_("Display remote resource(s)")), 
-    C1M(delete, N_("delete remote..."), N_("Delete non-collection resource(s)")),
-    C1M(rmcol, N_("rmcol remote..."), N_("Delete remote collections and ALL contents")),
-    C2M(copy, N_("copy source... dest"), N_("Copy resource(s) from source to dest")), 
-    C2M(move, N_("move source... dest"), N_("Move resource(s) from source to dest")),
-    
-/* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND */
-    
-    C1(lock, N_("lock resource"), N_("Lock given resource")),
-    C1(unlock, N_("unlock resource"), N_("Unlock given resource")),
-    C1(discover, N_("discover resource"), N_("Display lock information for resource")),
-    C1(steal, N_("steal resource"), N_("Steal lock token for resource")),
-    { cmd_showlocks, true, 0, 0, parmscope_none, execute_showlocks,
-      "showlocks", N_("Display list of owned locks") },
-    
-#if 0
-    C1(propedit, "propedit resource", "Enter property editor for resource"),
-#endif
-    C1(propnames, "propnames res", "Names of properties defined on resource"),
-
-    { cmd_chexec, true, 2, 2, parmscope_none, execute_chexec,
-      N_("chexec [+|-] remote"), N_("Change isexecutable property of resource") },
-    
-    { cmd_propget, true, 1, 2, parmscope_none, execute_propget,
-      N_("propget res [propname]"), 
-      N_("Retrieve properties of resource") },
-    { cmd_propdel, true, 2, 2, parmscope_none, execute_propdel,
-      N_("propdel res propname"), 
-      N_("Delete property from resource") },
-    { cmd_propset, true, 3, 3, parmscope_none, execute_propset,
-      N_("propset res propname value"),
-      N_("Set property on resource") },
-
-    { cmd_set, false, 0, 2, parmscope_none, execute_set, 
-      N_("set [option] [value]"), N_("Set an option, or display options") },
-    { cmd_open, false, 1, 1, parmscope_none, open_connection, 
-      "open URL", N_("Open connection to given URL") },
-    { cmd_close, true, 0, 0, parmscope_none, close_connection, 
-      "close", N_("Close current connection") },
-    { cmd_echo, false, 1, CMD_VARY, parmscope_remote, execute_echo, 
-      "echo", NULL },
-    { cmd_quit, false, 0, 1, parmscope_none, NULL, "quit", N_("Exit program") },
-    /* Unconnected operation, 1 mandatory argument */
-    { cmd_unset, false, 1, 2, parmscope_none, execute_unset, 
-      N_("unset [option] [value]"), N_("Unsets or clears value from option.") },
-    /* Unconnected operation, 0 arguments */
-    UO1(lcd, N_("lcd [directory]"), N_("Change local working directory")), 
-    { cmd_lls, false, 0, CMD_VARY, parmscope_local, execute_lls, 
-      N_("lls [options]"), N_("Display local directory listing") },
-    U0(lpwd, N_("Print local working directory")),
-    { cmd_logout, true, 0, 0, parmscope_none, execute_logout, "logout",
-      N_("Logout of authentication session") },
-    UO1(help, N_("help [command]"), N_("Display help message")), 
-    U0(version, NULL),
-    { cmd_unknown, 0 } /* end-of-list marker, DO NOT move */
-
-/* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND. */
-
-#undef C1
-#undef U0
-#undef UO1
-#undef C2M
-#undef C1M
-};    
-
 static void usage(void)
 {
     printf(_(
@@ -235,20 +139,25 @@ static void usage(void)
 "Please send bug reports and feature requests to <cadaver@webdav.org>\n"), progname);
 }
 
-static void execute_version(void)
+static void init_locking(void)
 {
-    printf(PACKAGE " " VERSION "\n%s\n", ne_version_string());
+    lock_store_fn = get_option(opt_lockstore);
+    lock_store = ne_lockstore_create();
+    /* TODO: read in lock list from ~/.davlocks */
 }
 
-static void close_connection(void)
+static void finish_locking(void)
+{
+    /* TODO: write out lock list to ~/.davlocks */
+}
+
+void close_connection(void)
 {
-    finish_locking(lockstore);
     ne_session_destroy(session);
-    sock_destroy_ssl_context(ssl_context);
     have_connection = false;
     NE_FREE(path);
     NE_FREE(old_path);
-    printf(_("Connection to `%s' closed.\n"), server_hostname);
+    printf(_("Connection to `%s' closed.\n"), server.host);
 }
 
 /* Sets the current collection to the given path.  Returns zero on
@@ -273,30 +182,14 @@ int set_path(const char *newpath)
     }
 }
 
-static void redirect_notify(void *userdata, const char *src, const char *dest)
-{
-    switch (out_state) {
-    case out_none:
-	printf("Redirected to %s\n", dest);
-	break;
-    case out_incommand:
-	printf("redirected to %s...", dest);
-	break;
-    case out_transfer_start:
-    case out_transfer_plain:
-    case out_transfer_pretty:
-	/* FIXME */
-	break;
-    }
-}
-
-static int privkey_prompt(void *userdata, const char *filename, 
-			  char *buf, int buflen)
+static int privkey_prompt(void *userdata, char *buf, size_t buflen)
 {
     static char prompt[BUFSIZ];
+    const char *filename = userdata;
     int len;
 
-    snprintf(prompt, BUFSIZ, _("Enter private key password for `%s': "), filename);
+    snprintf(prompt, BUFSIZ, _("Enter private key password for `%s': "), 
+	     filename);
     
     do {
 	char *pass = fm_getpassword(prompt);
@@ -315,31 +208,57 @@ static int privkey_prompt(void *userdata, const char *filename,
     return 0;
 }
 
+static int cert_verify(void *ud, int failures, const ne_ssl_certificate *c)
+{
+    char *tmp;
+
+    printf(_("WARNING: Untrusted server certificate presented:\n"));
+    if (failures & NE_SSL_CNMISMATCH) {
+	printf(_("Certificate was issued to hostname `%s' rather than `%s'\n"),
+	       c->subject->commonName, server.host);
+	printf(_("This connection could have been intercepted.\n"));
+    }
+
+#define PRINT_AND_FREE(str, dn) \
+tmp = ne_ssl_readable_dname(dn); printf(str, tmp); free(tmp)
+
+    PRINT_AND_FREE(_("Issued to: %s\n"), c->subject);
+    PRINT_AND_FREE(_("Issued by: %s\n"), c->subject);
+
+    if (isatty(STDIN_FILENO)) {
+	printf(_("Do you wish to accept the certificate? (y/n) "));
+	return !yesno();
+    } else {
+	printf(_("Certificate rejected.\n"));
+	return -1;
+    }
+}
+
 static int setup_ssl(void)
 {
     char *cc = get_option(opt_cert);
     char *key = get_option(opt_certkey);
-    
-    if (ne_set_secure(session, 1)) {
-	printf(
-	    _("SSL support has not be compiled into this application.\n"
-	      "URLs beginning with \"https:\" cannot be used!\n"));
-	return -1;
-    } 
-    
+
+    if (ne_ssl_load_default_ca(session)) {
+	printf(_("Warning: Could not load default CA list from SSL library.\n"
+		 "Check SSL library is installed correctly.\n"));
+    }
+	      
+
+    ne_ssl_set_verify(session, cert_verify, NULL);
+
     /* Set the client cert too if we have one. */
     if (cc != NULL) {
 	
 	/* Use our private key password prompt. */
-	sock_set_key_prompt(ssl_context, privkey_prompt, NULL);
+	ne_ssl_keypw_prompt(session, privkey_prompt, cc);
 
-	if (sock_set_client_cert(ssl_context, cc, key)) {
+	if (ne_ssl_load_pem(session, cc, key)) {
 	    printf(_("Could not use client certificate `%s' (wrong password)?\n"),
 		   cc);
 	    return -1;
 	} else {
 	    printf(_("Using client certificate `%s'.\n"), cc);
-	    ne_set_secure_context(session, ssl_context);
 	}
     }
 
@@ -347,12 +266,11 @@ static int setup_ssl(void)
 }
 
 /* FIXME: Leaky as a bucket */
-static void open_connection(const char *url)
+void open_connection(const char *url)
 {
     char *proxy_host = get_option(opt_proxy), *pnt;
     ne_server_capabilities caps;
-    struct uri uri = {0};
-    int ret;
+    int ret, use_ssl = 0;
 
     if (have_connection) {
 	close_connection();
@@ -361,81 +279,78 @@ static void open_connection(const char *url)
 	NE_FREE(old_path);
     }
 
-    session = ne_session_create();
-    ssl_context = sock_create_ssl_context();
-    ne_redirect_register(session, NULL, redirect_notify, NULL);
+    server.path = NULL; /* this is used as a temporary store. */
+    ne_uri_free(&server);
 
-    init_locking(lockstore);
-    
-    ne_set_progress(session, transfer_progress, NULL);
-    ne_set_status(session, connection_status, NULL);
-
-    NE_FREE(server_hostname);
-	
     /* Single argument: see whether we have a path or scheme */
     if (strchr(url, '/') == NULL) {
 	/* No path, no scheme -> just a hostname */
 	pnt = strchr(url, ':');
 	if (pnt != NULL) {
 	    *pnt++ = '\0';
-	    uri.port = atoi(pnt);
+	    server.port = atoi(pnt);
 	} else {
-	    uri.port = 80;
+	    server.port = 80;
 	}
-	uri.host = ne_strdup(url);
+	server.host = ne_strdup(url);
+	server.scheme = ne_strdup("http");
     } else {
 	/* Parse the URL */
-	if (uri_parse(url, &uri, NULL) || uri.host == NULL) {
+	if (ne_uri_parse(url, &server) || server.host == NULL) {
 	    printf(_("Could not parse URL `%s'\n"), url);
 	    return;
 	}
 
-	if (uri.scheme == NULL || strcasecmp(uri.scheme, "http") == 0) {
-	    if (uri.port == -1) {
-		uri.port = 80;
-	    }
-	} else if (strcasecmp(uri.scheme, "https") == 0) {
-	    if (uri.port == -1) {
-		uri.port = 443;
-	    }
-	    /* Set up SSL stuff. */
-	    if (setup_ssl()) {
+	if (server.scheme == NULL)
+	    server.scheme = ne_strdup("http");
+
+	if (!server.port)
+	    server.port = ne_uri_defaultport(server.scheme);
+
+	if (strcasecmp(server.scheme, "https") == 0) {
+	    if (!ne_supports_ssl()) {
+		printf(_("SSL is not enabled.\n"));
 		return;
 	    }
-	} else {
-	    printf(_("Unrecognized URL scheme `%s'.\n"), uri.scheme);
-	    uri_free(&uri);
-	    return;
+	    use_ssl = 1;
 	}
     }
+    
+    session = ne_session_create(server.scheme, server.host, server.port);
 
-    if (uri.path == NULL) {
-	uri.path = ne_strdup("/");
+    if (use_ssl && setup_ssl()) {
+	return;
+    }
+
+    ne_lockstore_register(lock_store, session);
+    
+    ne_set_progress(session, transfer_progress, NULL);
+    ne_set_status(session, connection_status, NULL);
+
+    if (server.path == NULL) {
+	server.path = ne_strdup("/");
     } else {
-	if (!uri_has_trailing_slash(uri.path)) {
-	    CONCAT2(pnt, uri.path, "/");
-	    free(uri.path);
-	    uri.path = pnt;
+	if (!ne_path_has_trailing_slash(server.path)) {
+	    pnt = ne_concat(server.path, "/", NULL);
+	    free(server.path);
+	    server.path = pnt;
 	}
     }
 
     /* Get the proxy details */
     if (proxy_host != NULL) {
 	if (get_option(opt_proxy_port) != NULL) {
-	    proxy_port = atoi((char *)get_option(opt_proxy_port));
+	    proxy_port = atoi(get_option(opt_proxy_port));
 	} else {
 	    proxy_port = 8080;
 	}
 	proxy_hostname = proxy_host;
     }
 
-    server_hostname = uri.host;
-    server_port = uri.port;
-
 #ifdef ENABLE_NETRC
     {
 	netrc_entry *found;
-	found = search_netrc(netrc_list, server_hostname);
+	found = search_netrc(netrc_list, server.host);
 	if (found != NULL) {
 	    if (found->account && found->password) {
 		server_username = found->account;
@@ -445,54 +360,39 @@ static void open_connection(const char *url)
     }
 #endif /* ENABLE_NETRC */
     have_connection = false;
-    /* Default to '/' if no path given */
-#if 0
-    if (open_path != NULL) {
-	if (uri_has_trailing_slash(open_path)) {
-	    use_path = ne_strdup(open_path);
-	} else if (open_path[0] == '/') {
-	    CONCAT2(use_path, open_path, "/");
-	} else {
-	    CONCAT3(use_path, "/", open_path, "/");
-	}
-    } else {
-	use_path = ne_strdup("/");
-    }
-#endif
 
     ne_set_expect100(session, use_expect);
     ne_set_useragent(session, PACKAGE "/" VERSION);
     ne_set_server_auth(session, supply_creds_server, NULL);
     ne_set_proxy_auth(session, supply_creds_proxy, NULL);
     
-    /* FIXME: returns here LEAK */
     if (proxy_host) {
-	if (ne_session_proxy(session, proxy_hostname, proxy_port) != NE_OK) {
-	    printf(_("Could not resolve proxy server hostname `%s'.\n"),
-		    proxy_hostname);
-	    return;
-	}
+	ne_session_proxy(session, proxy_hostname, proxy_port);
     }
 
-    if (ne_session_server(session, server_hostname, server_port) != NE_OK) {
-	printf(_("Could not resolve server hostname `%s'.\n"), server_hostname);
-	return;
-    }
-    
-    ret = ne_options(session, uri.path, &caps);
+    ret = ne_options(session, server.path, &caps);
     
     switch (ret) {
     case NE_OK:
 	have_connection = true;
 	path = NULL;
-	if (set_path(uri.path)) {
+	if (set_path(server.path)) {
 	    close_connection();
 	} else {
-	    path = uri.path;
+	    path = server.path;
 	}
 	break;
     case NE_CONNECT:
-	printf(_("Could not connect to remote host.\n"));
+	if (proxy_host) {
+	    printf(_("Could not connect to `%s' on port %d:\n%s\n"),
+		   proxy_hostname, proxy_port, ne_get_error(session));
+	} else {
+	    printf(_("Could not connect to `%s' on port %d:\n%s\n"),
+		   server.host, server.port, ne_get_error(session));
+	}
+	break;
+    case NE_LOOKUP:
+	puts(ne_get_error(session));
 	break;
     default:
 	/* FIXME: This is NOT a "could not open connection" error */
@@ -545,7 +445,7 @@ static void parse_args(int argc, char **argv)
 #ifdef HAVE_ADD_HISTORY
 	{ 
 	    char *run_cmd;
-	    CONCAT2(run_cmd, "open ", argv[optind]);
+	    run_cmd = ne_concat("open ", argv[optind], NULL);
 	    add_history(run_cmd);
 	    free(run_cmd);
 	}
@@ -610,28 +510,23 @@ static int execute_command(const char *line)
 	ret = -1;
     } else {
 	/* Cast away */
-	void (*take0)(void)=cmd->handler, 
-	    (*take1)(const char *)=cmd->handler, 
-	    (*take2)(const char *, const char *)=cmd->handler,
-	    (*take3)(const char *, const char *, const char *)=cmd->handler,
-	    (*takeV)(int, const char **)=cmd->handler;
 	/* with a nod in the general direction of apache */
 	switch (cmd->max_args) {
-	case 0: (*take0)(); break;
+	case 0: cmd->handler.take0(); break;
 	case 1: /* tokens[1]==NULL if argcount==0 */
-	    (*take1)(tokens[1]); break; 
+	    cmd->handler.take1(tokens[1]); break; 
 	case 2: 
 	    if (argcount <=1) {
-		(*take2)(tokens[1], NULL);
+		cmd->handler.take2(tokens[1], NULL);
 	    } else {
-		(*take2)(tokens[1], tokens[2]);
+		cmd->handler.take2(tokens[1], tokens[2]);
 	    }
 	    break;
 	case 3:
-	    (*take3)(tokens[1], tokens[2], tokens[3]);
+	    cmd->handler.take3(tokens[1], tokens[2], tokens[3]);
 	    break;
 	case CMD_VARY:
-	    (*takeV)(argcount, (const char **) &tokens[1]);
+	    cmd->handler.takeV(argcount, (const char **) &tokens[1]);
 	default:
 	    break;
 	}
@@ -660,7 +555,6 @@ static RETSIGTYPE quit_handler(int sig)
 
 void init_signals(void)
 {
-    signal(SIGPIPE, SIG_IGN);
     signal(SIGTERM, quit_handler);
     signal(SIGABRT, quit_handler);
     signal(SIGQUIT, quit_handler);
@@ -670,8 +564,7 @@ void init_signals(void)
 static void init_netrc(void)
 {
 #ifdef ENABLE_NETRC
-    char *netrc;
-    CONCAT2(netrc, getenv("HOME"), "/.netrc");
+    char *netrc = ne_concat(getenv("HOME"), "/.netrc", NULL);
     netrc_list = parse_netrc(netrc);
 #endif
 }
@@ -681,7 +574,7 @@ static void init_rcfile(void)
     char *rcfile, buf[BUFSIZ];
     struct stat st;
     FILE *f;
-    CONCAT2(rcfile, getenv("HOME"), "/.cadaverrc");
+    rcfile = ne_concat(getenv("HOME"), "/.cadaverrc", NULL);
     if (stat(rcfile, &st) != 0) {
 	NE_DEBUG(DEBUG_FILES, "No rcfile.\n");
     } else {
@@ -692,8 +585,7 @@ static void init_rcfile(void)
 	} else {
 	    for (;;) {
 		if (fgets(buf, BUFSIZ, f) != NULL) {
-		    STRIP_EOL(buf);
-		    execute_command(buf);
+		    execute_command(ne_shave(buf, "\r\n"));
 		} else {
 		    break;
 		}
@@ -765,7 +657,7 @@ static char *remote_completion(const char *text, int state)
     while (current) {
 	/* Massage the absolute URI to a URI relative to our path */
 	/* Copy & paste & search & replace from ls.c */
-	if (uri_has_trailing_slash(current->uri)) {
+	if (ne_path_has_trailing_slash(current->uri)) {
 	    current->uri[strlen(current->uri)-1] = '\0';
 	}
 
@@ -775,11 +667,14 @@ static char *remote_completion(const char *text, int state)
 	} else {
 	    name = current->uri;
 	}
-	name = uri_unescape(name);
+	name = ne_path_unescape(name);
 
 	if (strncmp(text, name, len) == 0) {
 	    current = current->next;
-	    return ne_strdup(name);
+	    /* FIXME: readline docs say readline will free() this when
+	     * it's finished with, although 'memprof' shows that it
+	     * leaks. */
+	    return name;
 	}
 
 	current = current->next;
@@ -813,7 +708,9 @@ static char **completion(const char *text, int start, int end)
 		/* TODO */
 		break;
 	    case parmscope_remote:
-		matches = rl_completion_matches(text, remote_completion);
+		if (have_connection) {
+		    matches = rl_completion_matches(text, remote_completion);
+		}
 		break;
 	    }
 	}
@@ -872,8 +769,7 @@ char *readline(const char *prompt)
     }
     ret = fgets(buf, 256, stdin);
     if (ret) {
-	STRIP_EOL(buf);
-	return ne_strdup(buf);
+	return ne_strdup(ne_shave(buf, "\r\n"));
     } else {
 	return NULL;
     }
@@ -887,7 +783,7 @@ static void init_options(void)
     
     if (user && hostname) {
 	/* set this here so they can override it */
-	CONCAT4(lockowner, "mailto:", user, "@", hostname);
+	lockowner = ne_concat("mailto:", user, "@", hostname, NULL);
 	set_option(opt_lockowner, lockowner);
     } else {
 	set_option(opt_lockowner, NULL);
@@ -896,8 +792,9 @@ static void init_options(void)
     set_option(opt_editor, NULL);
     set_option(opt_namespace, ne_strdup(DEFAULT_NAMESPACE));
     set_bool_option(opt_overwrite, 1);
+    set_bool_option(opt_quiet, 1);
     lockdepth = NE_DEPTH_INFINITE;
-    lockstore = ne_lockscope_exclusive;
+    lockscope = ne_lockscope_exclusive;
 
     /* This is what Markus Kahn says we should do. */
     if ((tmp = getenv("LC_ALL")) ||
@@ -914,7 +811,7 @@ static void init_options(void)
 int main(int argc, char *argv[])
 {
     int ret = 0;
-    char *home = getenv("HOME");
+    char *home = getenv("HOME"), *tmp;
 
     path = NULL;
     have_connection = false;
@@ -932,18 +829,18 @@ int main(int argc, char *argv[])
 	return -1;
     }
 
-    sock_init();
-
-    CONCAT2(lockstore, home, "/.davlocks");
+    ne_sock_init();
 
     /* Options before rcfile, so rcfile settings can
      * override defaults */
+    tmp = ne_concat(home, "/.cadaver-locks", NULL);
+    set_option(opt_lockstore, tmp);
     init_options();
     init_netrc();
 
     init_signals();
-
     init_rcfile();
+    init_locking();
     
     parse_args(argc, argv);
 
@@ -965,10 +862,14 @@ int main(int argc, char *argv[])
 	    free(cmd);
 	}
     }
+
     if (have_connection) {
 	close_connection();
     }
-    finish_locking(lockstore);
+
+    finish_locking();
+
+    ne_sock_exit();
 
     return 0;
 }
@@ -1200,7 +1101,7 @@ static int supply_creds_server(void *userdata, const char *realm, int attempt,
 
     return supply_creds(
 	_("Authentication required for %s on server `%s':\n"), realm,
-	server_hostname, username, password);
+	server.host, username, password);
 }
 
 static int supply_creds_proxy(void *userdata, const char *realm, int attempt,
