@@ -77,8 +77,6 @@
 #include "options.h"
 #include "utils.h"
 
-int yesno(void);
-
 /* Local variables */
 int child_running; /* true when we have a child running */
 
@@ -107,6 +105,8 @@ C(propnames), C(edit),
     { cmd_unknown, NULL }
 };    
 
+extern const struct command commands[]; /* prototype */
+
 /* Tell them we are doing 'VERB' to 'NOUN'.
  * (really 'METHOD' to 'RESOURCE' but hey.) */
 void out_start(const char *verb, const char *noun)
@@ -126,7 +126,7 @@ void out_result(int ret)
 	out_success();
 	break;
     case NE_AUTH:
-    case NE_AUTHPROXY:
+    case NE_PROXYAUTH:
 	output(o_finish, _("authentication failed.\n"));
 	break;
     case NE_CONNECT:
@@ -135,10 +135,6 @@ void out_result(int ret)
     case NE_TIMEOUT:
 	output(o_finish, _("connection timed out.\n"));
 	break;
-    case NE_REDIRECT:
-	output(o_finish, _("redirected to a different server:\n%s\n"),
-	       ne_redirect_location(session));
-	break;
     default:
 	output(o_finish, _("failed:\n%s\n"), ne_get_error(session));
 	break;
@@ -152,42 +148,6 @@ int out_handle(int ret)
 }
 
 /* The actual commands */
-
-void init_locking(const char *lockstore)
-{
-    lock_session = ne_lock_register(session);
-    /* TODO: read in lock list from ~/.davlocks */
-}
-
-void finish_locking(const char *lockstore)
-{
-    /* TODO: write out lock list to ~/.davlocks */
-}
-
-/* Auxiliaries */
-
-static void 
-map_multi(void (*func)(const char *), int argc, const char *argv[]);
-
-static void simple_copy(const char *from, const char *to);
-static void simple_put(const char *local, const char *remote);
-static void simple_move(const char *src, const char *dest);
-
-static void do_copymove(int argc, const char *argv[],
-			const char *v1, const char *v2,
-			void (*cb)(const char *, const char *));
-
-static char *clever_path(const char *p, const char *src, 
-			 const char *dest);
-
-static const char *choose_pager(void);
-
-static void dispatch(const char *verb, const char *filename, 
-		     int (*func)(ne_session *, const char *), 
-		     const char *pass);
-
-static void display_help_message(void);
-
 #ifdef HAVE_LIBREADLINE
 
 /* Command name generator for readline.
@@ -213,7 +173,7 @@ char *command_generator(const char *text, int state)
 
 #endif
 
-void execute_logout(void)
+static void execute_logout(void)
 {
     ne_forget_auth(session);
 }
@@ -234,42 +194,6 @@ const struct command *get_command(const char *name)
 }
 
 
-/* A bit like Haskell's map function... applies func to each
- * of the first argc items in argv */
-static void 
-map_multi(void (*func)(const char *), int argc, const char *argv[])
-{
-    int n;
-    for(n = 0; n < argc; n++) {
-	(*func)(argv[n]);
-    }
-}
-
-void multi_mkcol(int argc, const char *argv[])
-{
-    map_multi(execute_mkcol, argc, argv);
-}
-
-void multi_delete(int argc, const char *argv[])
-{
-    map_multi(execute_delete, argc, argv);
-}
-
-void multi_rmcol(int argc, const char *argv[])
-{
-    map_multi(execute_rmcol, argc, argv);
-}
-
-void multi_less(int argc, const char *argv[])
-{
-    map_multi(execute_less, argc, argv);
-}
-
-void multi_cat(int argc, const char *argv[])
-{
-    map_multi(execute_cat, argc, argv);
-}
-
 static void dispatch(const char *verb, const char *filename, 
 		     int (*func)(ne_session *, const char *), const char *arg)
 {
@@ -277,35 +201,78 @@ static void dispatch(const char *verb, const char *filename,
     out_result((*func)(session, arg));
 }
 
-static char * getowner(void)
+char *getowner(void)
 {
-#if 0
-    /* mod_dav breaks if you submit an owner */
-    return ne_strdup("");
-#else
     char *ret, *owner = get_option(opt_lockowner);
     if (owner) {
-	CONCAT3(ret, "<href>", owner, "</href>");
+	ret = ne_concat("<href>", owner, "</href>", NULL);
 	return ret;
     } else {
 	return NULL;
     }
-#endif
 }
 
-/* If in UTF-8 mode, simply duplicates 'str'.
- * When not in UTF-8 mode, presume 'str' is ISO-8859-1,
- * and UTF-8 encode it.
- *
- * TODO: I'm not 100% sure this is right. At least, the strdup
- * doesn't cope with embedded NUL's which it should, right?
- */
+/* FIXME: joe: these are really wrong and result from my lack of
+ * understanding of character set issues.  I think this needs iconv(),
+ * but iconv isn't really portable, so what's the fallback for
+ * platforms without iconv? Reject any non-ASCII input?  And even if
+ * iconv is present, how is the input charset encoding determined? */
+static char *cad_utf8_encode(const char *str)
+{
+    char *buffer = ne_malloc(strlen(str) * 2 + 1);
+    int in, len = strlen(str);
+    char *out;
+
+    for (in = 0, out = buffer; in < len; in++, out++) {
+	if ((unsigned char)str[in] <= 0x7D) {
+	    *out = str[in];
+	} else {
+	    *out++ = 0xC0 | ((str[in] & 0xFC) >> 6);
+	    *out = str[in] & 0xBF;
+	}
+    }
+
+    /* nul-terminate */
+    *out = '\0';
+    return buffer;
+}
+
+/* Single byte range 0x00 -> 0x7F */
+#define SINGLEBYTE_UTF8(ch) (((unsigned char) (ch)) < 0x80)
+
+static char *cad_utf8_decode(const char *str)
+{
+    int n, m, len = strlen(str);
+    char *dest = ne_malloc(len + 1);
+    
+    for (n = 0, m = 0; n < len; n++, m++) {
+	if (SINGLEBYTE_UTF8(str[n])) {
+	    dest[m] = str[n];
+	} else {
+	    /* This just deals with 8-bit data, which will be encoded
+	     * as two bytes of UTF-8 */
+	    if ((len - n < 2) || (str[n] & 0xFC) != 0xC0) {
+		free(dest);
+		return NULL;
+	    } else {
+		dest[m] = ((str[n] & 0x03) << 6) | (str[n+1] & 0x3F);
+		n++;
+	    }
+	}
+    }
+    
+    dest[m] = '\0';
+    return dest;
+}
+
+/* If in UTF-8 mode, simply duplicates 'str'.  When not in UTF-8 mode,
+ * presume 'str' is ISO-8859-1, and UTF-8 encode it. */
 static char *utf8_encode(const char *str)
 {
     if (get_bool_option(opt_utf8)) {
 	return ne_strdup(str);
     } else {
-	return ne_utf8_encode(str);
+	return cad_utf8_encode(str);
     }
 }
 
@@ -323,7 +290,7 @@ static char *utf8_decode(const char *str)
     if (get_bool_option(opt_utf8)) {
 	return ne_strdup(str);
     } else {
-	return ne_utf8_decode(str);
+	return cad_utf8_decode(str);
     }
 }
 
@@ -336,9 +303,8 @@ static char *true_path(const char *res)
     char *full;
     full = resolve_path(path, res, false);
     if (getrestype(full) == resr_collection) {
-	if (!uri_has_trailing_slash(full)) {
-	    char *tmp;
-	    CONCAT2(tmp, full, "/");
+	if (!ne_path_has_trailing_slash(full)) {
+	    char *tmp = ne_concat(full, "/", NULL);
 	    free(full);
 	    full = tmp;
 	}
@@ -391,15 +357,25 @@ static const char *get_depth(int d)
     }
 }
 
+static void print_uri(const ne_uri *uri)
+{
+    char *str = ne_uri_unparse(uri);
+    printf("%s", str);
+    free(str);
+}
+
 static void print_lock(const struct ne_lock *lock)
 {
-    printf(_("Lock <%s> on %s:\n"
-	     "  Scope: %s  Type: %s  Timeout: %s  Depth: %s\n"
+    char *uri = ne_uri_unparse(&lock->uri);
+    printf(_("Lock token <%s>:\n"
+	     "  Depth %s on `%s'\n"
+	     "  Scope: %s  Type: %s  Timeout: %s\n"
 	     "  Owner: %s\n"), 
-	       lock->token, lock->uri,
+	   lock->token, get_depth(lock->depth), uri,
 	   get_lockscope(lock->scope),
 	   get_locktype(lock->type), get_timeout(lock->timeout),
-	   get_depth(lock->depth), lock->owner?lock->owner:_("(none)"));
+	   lock->owner?lock->owner:_("(none)"));
+    free(uri);
 }
 
 static void discover_result(void *userdata, const struct ne_lock *lock, 
@@ -418,26 +394,6 @@ static void discover_result(void *userdata, const struct ne_lock *lock,
     }
 }
 
-void execute_discover(const char *res)
-{
-    char *real_remote;
-    int ret, count = 0;
-    real_remote = resolve_path(path, res, false);
-    out_start(_("Discovering locks on"), res);
-    ret = ne_lock_discover(session, real_remote, discover_result, &count);
-    switch (ret) {
-    case NE_OK:
-	if (count == 0) {
-	    printf(" no locks found.\n");
-	}
-	break;
-    default:
-	out_result(ret);
-	break;
-    }
-    free(real_remote);
-}
-
 static void steal_result(void *userdata, const struct ne_lock *lock, 
 			 const char *uri, const ne_status *status)
 {
@@ -446,8 +402,9 @@ static void steal_result(void *userdata, const struct ne_lock *lock,
 	if (*count == 0) {
 	    printf("\n");
 	}
-	printf("%s: <%s>\n", lock->uri, lock->token);
-	ne_lock_add(lock_session, ne_lock_copy(lock));
+	print_uri(&lock->uri);
+	printf(": <%s>\n", lock->token);
+	ne_lockstore_add(lock_store, ne_lock_copy(lock));
 	*count += 1;
     } else {
 	printf(_("Failed on %s: %d %s\n"), uri,
@@ -455,13 +412,14 @@ static void steal_result(void *userdata, const struct ne_lock *lock,
     }
 }
 
-void execute_steal(const char *res)
+static void do_discover(const char *res, const char *mesg,
+			ne_lock_result result_cb)
 {
     char *real_remote;
     int ret, count = 0;
     real_remote = resolve_path(path, res, false);
-    out_start(_("Stealing locks on"), res);
-    ret = ne_lock_discover(session, real_remote, steal_result, &count);
+    out_start(mesg, res);
+    ret = ne_lock_discover(session, real_remote, result_cb, &count);
     switch (ret) {
     case NE_OK:
 	if (count == 0) {
@@ -472,24 +430,35 @@ void execute_steal(const char *res)
 	out_result(ret);
 	break;
     }
-    
     free(real_remote);
 }
 
-static void showlocks_walker(struct ne_lock *lock, void *userdata)
+static void execute_discover(const char *res)
 {
-    print_lock(lock);
+    do_discover(res, _("Discovering locks on"), discover_result);
 }
 
-void execute_showlocks(void)
+static void execute_steal(const char *res)
 {
-    printf(_("List of owned locks:\n"));
-    if (ne_lock_iterate(lock_session, showlocks_walker, NULL) == 0) {
+    do_discover(res, _("Stealing locks on"), steal_result);
+}
+
+static void execute_showlocks(void)
+{
+    int count = 0;
+    struct ne_lock *lk;
+    
+    for (lk = ne_lockstore_first(lock_store); lk != NULL;
+	 lk = ne_lockstore_next(lock_store), count++) {
+        print_lock(lk);
+    }
+
+    if (count == 0) {
 	printf(_("No owned locks.\n"));
     }
 }
 
-void execute_lock(const char *res)
+static void execute_lock(const char *res)
 {
     char *real_remote;
     struct ne_lock *lock;
@@ -498,9 +467,8 @@ void execute_lock(const char *res)
     real_remote = resolve_path(path, res, false);
     rtype = getrestype(real_remote);
     if (rtype == resr_collection) {
-	if (!uri_has_trailing_slash(real_remote)) {
-	    char *tmp;
-	    CONCAT2(tmp, real_remote, "/");
+	if (!ne_path_has_trailing_slash(real_remote)) {
+	    char *tmp = ne_concat(real_remote, "/", NULL);
 	    free(real_remote);
 	    real_remote = tmp;
 	}
@@ -508,8 +476,7 @@ void execute_lock(const char *res)
     } else {
 	out_start(_("Locking"), res);
     }
-    lock = ne_calloc(sizeof *lock);
-    lock->type = ne_locktype_write;
+    lock = ne_lock_create();
     lock->scope = lockscope;
     lock->owner = getowner();
     if (rtype == resr_normal) {
@@ -519,44 +486,48 @@ void execute_lock(const char *res)
 	/* use value of lockdepth option. */
 	lock->depth = lockdepth;
     }
-    lock->uri = real_remote;
-    if (out_handle(ne_lock(session,lock))) {
-	/* Locktoken: <%s>\n", lock->token); */
-	ne_lock_add(lock_session, lock);
+    ne_fill_server_uri(session, &lock->uri);
+    lock->uri.path = ne_strdup(real_remote);
+    if (out_handle(ne_lock(session, lock))) {
+	/* success: remember the lock. */
+	ne_lockstore_add(lock_store, lock);
     } else {
-	ne_lock_free(lock);
+	/* otherwise, throw it away */
+	ne_lock_destroy(lock);
     }
+    free(real_remote);
 }
 
-void execute_unlock(const char *res)
+static void execute_unlock(const char *res)
 {
     struct ne_lock *lock;
-    char *real_remote;
-    int in_list = 1;
+    char *real_remote = true_path(res);
+
     out_start(_("Unlocking"), res);
-    real_remote = true_path(res);
-    lock = ne_lock_find(lock_session, real_remote);
+    /* use server URI as temp store. */
+    server.path = real_remote;
+    lock = ne_lockstore_findbyuri(lock_store, &server);
     if (!lock) {
-	lock = ne_calloc(sizeof *lock);
+	lock = ne_lock_create();
 	lock->token = readline(_("Enter locktoken: "));
 	if (!lock->token || strlen(lock->token) == 0) {
-	    free(lock);
-	    return;
+	    goto unlock_fail;
 	}
-	lock->uri = real_remote;
-	in_list = 0;
+	ne_fill_server_uri(session, &lock->uri);
+	lock->uri.path = ne_strdup(real_remote);
+    } else {
+	/* remove the lock from the lockstore */
+	ne_lockstore_remove(lock_store, lock);
     }
+
     out_result(ne_unlock(session, lock));
 
-    if (in_list) {
-	ne_lock_remove(lock_session, lock);
-	free(real_remote);
-    } 
-    /* else real_remote will get free'd here... */
-    ne_lock_free(lock);
+unlock_fail:
+    free(real_remote);
+    ne_lock_destroy(lock);
 }
 
-void execute_mkcol(const char *filename)
+static void execute_mkcol(const char *filename)
 {
     char *remote;
     remote = resolve_path(path, filename, true);
@@ -647,17 +618,17 @@ static void propop(const char *descr, const char *res,
 }    
 
 
-void execute_propset(const char *res, const char *name, const char *value)
+static void execute_propset(const char *res, const char *name, const char *value)
 {
     propop(_("Setting property on"), res, name, value);
 }
 
-void execute_propdel(const char *res, const char *name)
+static void execute_propdel(const char *res, const char *name)
 {
     propop(_("Deleting property on"), res, name, NULL);
 }
 
-void execute_propget(const char *res, const char *name)
+static void execute_propget(const char *res, const char *name)
 {
     ne_propname pnames[2] = {{NULL}, {NULL}};
     char *remote = resolve_path(path, res, false), *encname = NULL;
@@ -699,7 +670,7 @@ static void propname_results(void *userdata, const char *href,
     ne_propset_iterate(pset, propname_iterator, NULL);
 }
 
-void execute_propnames(const char *res)
+static void execute_propnames(const char *res)
 {
     char *remote;
     remote = resolve_path(path, res, false);
@@ -710,18 +681,26 @@ void execute_propnames(const char *res)
     free(remote);
 }
 
-static void remove_locks(const char *uri, int depth)
+static void remove_locks(const char *p, int depth)
 {
-    struct ne_lock *lck;
+    struct ne_lock *lk;
+    ne_uri sought;
     
-    lck = ne_lock_find(lock_session, uri);
-    if (lck != NULL) {
-	ne_lock_remove(lock_session, lck);
-    }
+    memset(&sought, 0, sizeof(sought));
+    ne_fill_server_uri(session, &sought);
+    sought.path = ne_strdup(p);
 
+    do {
+	lk = ne_lockstore_findbyuri(lock_store, &sought);
+	if (lk) {
+	    ne_lockstore_remove(lock_store, lk);
+	}
+    } while (lk);
+
+    ne_uri_free(&sought);
 }
 
-void execute_delete(const char *filename)
+static void execute_delete(const char *filename)
 {
     char *remote = resolve_path(path, filename, false);
     out_start(_("Deleting"), filename);
@@ -732,14 +711,14 @@ _("is a collection resource.\n"
 "Use `rmcol %s' to delete this collection and ALL its contents.\n"), 
 filename);
     } else {
-	if (!out_handle(ne_delete(session, remote))) {
+	if (out_handle(ne_delete(session, remote))) {
 	    remove_locks(remote, 0);
 	}
     }
     free(remote);
 }
 
-void execute_rmcol(const char *filename)
+static void execute_rmcol(const char *filename)
 {
     char *remote;
     remote = resolve_path(path, filename, true);
@@ -768,15 +747,14 @@ static char *clever_path(const char *p, const char *src,
 	ret = resolve_path(p, base_name(src), false);
     } else if (strcmp(dest, "..") == 0) {
 	char *parent;
-	parent = uri_parent(p);
+	parent = ne_path_parent(p);
 	ret = resolve_path(parent, base_name(src), false);
 	free(parent);
     } else if (!src_is_coll && dest_is_coll) {
 	/* Moving a file to a collection... the destination should
 	 * be the basename of file concated with the collection. */
-	char *tmp;
-	tmp = resolve_path(p, dest, true);
-	CONCAT2(ret, tmp, base_name(src));
+	char *tmp = resolve_path(p, dest, true);
+	ret = ne_concat(tmp, base_name(src), NULL);
 	free(tmp);
     } else {
 	ret = resolve_path(p, dest, false);
@@ -784,18 +762,13 @@ static char *clever_path(const char *p, const char *src,
     return ret;
 }
 
-static void cat_callback(void *userdata, const char *buf, size_t len)
-{
-    FILE *f = userdata;
-    if (fwrite(buf, len, 1, f) < 0) {
-	perror("fwrite");
-    }
-}
-
 static const char *choose_pager(void)
 {
     struct stat st;
     char *tmp;
+    tmp = get_option(opt_pager);
+    if (tmp != NULL)
+	return tmp;
     /* TODO: add an autoconf for less/more? */
     tmp = getenv("PAGER");
     if (tmp != NULL) {
@@ -821,7 +794,7 @@ static void kill_pager(FILE *p)
     pclose(p);
 }
 
-void execute_less(const char *resource)
+static void execute_less(const char *resource)
 {
     char *real_res;
     const char *pager;
@@ -834,32 +807,23 @@ void execute_less(const char *resource)
 	printf(_("Error! Could not spawn pager `%s':\n%s\n"), pager,
 		 strerror(errno));
     } else {
+	int fd = fileno(p);
 	child_running = true;
-	ne_read_file(session, real_res, cat_callback, p);
+	ne_get(session, real_res, fd);
 	kill_pager(p); /* Blocks until the pager quits */
 	child_running = false;
     }
 }
 
-void execute_cat(const char *resource)
+static void execute_cat(const char *resource)
 {
     char *real_res = resolve_path(path, resource, false);
     printf(_("Displaying `%s':\n"), real_res);
-    if (ne_read_file(session, real_res, cat_callback, stdout) != NE_OK) {
+    if (ne_get(session, real_res, STDOUT_FILENO) != NE_OK) {
 	printf(_("Failed: %s\n"), ne_get_error(session));
     }
 }
 
-void multi_copy(int argc, const char *argv[])
-{
-    do_copymove(argc, argv, _("copying"), _("copy"), simple_copy);
-}
-
-void multi_move(int argc, const char *argv[])
-{
-    do_copymove(argc, argv, _("moving"), _("move"), simple_move);
-}
-
 static void do_copymove(int argc, const char *argv[],
 			const char *v1, const char *v2,
 			void (*cb)(const char *, const char *))
@@ -914,17 +878,9 @@ static void simple_copy(const char *src, const char *dest)
 /* TODO: can do utf-8 handling here. */
 static char *escape_path(const char *p)
 {
-    return uri_abspath_escape(p);
+    return ne_path_escape(p);
 }
 
-/* Returns absolute filename which is 'filename' relative to 
- * (absolute filename) 'path'. e.g.
- *    resolve_path("/dav/", "asda") == "/dav/asda"
- *    resolve_path("/dav/", "/asda") == "/asda"
- * Also removes '..' segments, e.g.
- *    resolve_path("/dav/foobar/", "../fish") == "/dav/fish"
- * If isdir is true, ensures the return value has a trailing slash.
- */
 char *resolve_path(const char *p, const char *filename, int isdir)
 {
     char *ret, *pnt;
@@ -935,12 +891,11 @@ char *resolve_path(const char *p, const char *filename, int isdir)
 	ret = escape_path(p);
     } else {
 	pnt = escape_path(filename);
-	CONCAT2(ret, p, pnt);
+	ret = ne_concat(p, pnt, NULL);
 	free(pnt);
     }
     if (isdir && ret[strlen(ret)-1] != '/') {
-	char *newret;
-	CONCAT2(newret, ret, "/");
+	char *newret = ne_concat(ret, "/", NULL);
 	free(ret);
 	ret = newret;
     }
@@ -968,7 +923,7 @@ char *resolve_path(const char *p, const char *filename, int isdir)
     return ret;    
 }
 
-void execute_get(const char *remote, const char *local)
+static void execute_get(const char *remote, const char *local)
 {
     char *filename, *real_remote;
     real_remote = resolve_path(path, remote, false);
@@ -976,9 +931,11 @@ void execute_get(const char *remote, const char *local)
 	struct stat st;
 	/* Choose an appropriate local filename */
 	if (stat(base_name(remote), &st) == 0) {
+	    char buf[BUFSIZ];
 	    /* File already exists... don't overwrite */
-	    printf(_("Enter local filename for `%s': "), real_remote);
-	    filename = readline(NULL);
+	    snprintf(buf, BUFSIZ, _("Enter local filename for `%s': "),
+		     real_remote);
+	    filename = readline(buf);
 	    if (filename == NULL) {
 		free(real_remote);
 		printf(_("cancelled.\n"));
@@ -1004,176 +961,6 @@ void execute_get(const char *remote, const char *local)
     free(filename);
 }
 
-static int run_editor(const char *filename)
-{
-    char editcmd[BUFSIZ];
-    const char *editor;
-    struct stat before_st, after_st;
-    editor = get_option(opt_editor);
-    if (editor == NULL) {
-	editor = getenv("EDITOR");
-	if (editor == NULL) {
-	    editor = "vi";
-	}
-    }
-    snprintf(editcmd, BUFSIZ, "%s %s", editor, filename);
-    if (stat(filename, &before_st)) {
-	printf(_("Could not stat file: %s\n"), strerror(errno));
-	return -1;
-    }
-    printf("Running editor: `%s'...\n", editcmd);
-    system(editcmd);
-    if (stat(filename, &after_st)) {
-	printf(_("Error! Could not examine temporary file: %s\n"), 
-		 strerror(errno));
-	return -1;
-    }
-    if (before_st.st_mtime == after_st.st_mtime) {
-	/* File not changed. */
-	printf(_("No changes were made.\n"));
-	return -1;
-    } else {
-	printf(_("Changes were made.\n"));
-	return 0;
-    }	
-}
-
-/* Returns true if resource at URI is lockable. */
-static int is_lockable(const char *uri)
-{
-    ne_server_capabilities caps = {0};
-
-    /* TODO: for the mo, just check we're on a Class 2 server.
-     * Should check supportedlock property really. */
-
-    if (ne_options(session, uri, &caps) != NE_OK) {
-	return 0;
-    }
-
-    return caps.dav_class2;
-}
-
-
-/* TODO: this is great big heap of steaming trout.  */
-/* TODO: does this work under cygwin? mkstemp() may or may not open
-   the file using O_BINARY, and then we *do* upload it using O_BINARY,
-   so maybe this will screw things up. Maybe we should fcntl it and
-   set O_BINARY, if that is allowed under cygwin? */
-void execute_edit(const char *remote)
-{
-    char *real_remote;
-    unsigned int can_lock; /* can we LOCK it? */
-    struct ne_lock *lock = NULL;
-    char fname[] = "/tmp/cadaver-edit-XXXXXX";
-    int fd;
-    
-    real_remote = resolve_path(path, remote, false);
-
-    /* Don't let them edit a collection, since PUT to a collection is
-     * bogus. Really we want to be able to fetch a "DefaultDocument"
-     * property, and edit on that instead: IIS does expose such a
-     * property. Would be a nice hack to add the support to mod_dav
-     * too. */
-    if (getrestype(real_remote) == resr_collection) {
-	printf(_("You cannot edit a collection resource (%s).\n"),
-	       real_remote);
-	goto edit_bail;
-    }
-
-    can_lock = is_lockable(real_remote);
-
-    fd = mkstemp(fname);
-    if (fd == -1) {
-	printf(_("Could not create temporary file %s:\n%s\n"), fname,
-	       strerror(errno));
-	goto edit_bail;
-    }
-
-    /* Sanity check on the file perms. */
-    if (fchmod(fd, 0600) == -1) {
-	printf(_("Could not set file permissions for %s:\n%s\n"), fname,
-	       strerror(errno));
-	goto edit_bail;
-    }
-   
-    if (can_lock) {
-	/* FIXME: copy'n'friggin'paste */
-	lock = ne_calloc(sizeof *lock);
-	lock->type = ne_locktype_write;
-	lock->scope = ne_lockscope_exclusive;
-	lock->owner = getowner();
-	lock->depth = 0;
-	lock->uri = ne_strdup(real_remote);
-	out_start("Locking", remote);
-	if (out_handle(ne_lock(session, lock))) {
-	    ne_lock_add(lock_session, lock);
-	} else {
-	    ne_lock_free(lock);
-	    goto edit_close;
-	    return;
-	}
-    } else {
-	/* TODO: HEAD and get the Etag/modtime */
-    }
-    /* FIXME: copy'n'friggin'paste. */
-    output(o_transfer, _("Downloading `%s' to %s"), real_remote, fname);
-
-    /* Don't puke if get fails -- perhaps we are creating a new one? */
-    out_result(ne_get(session, real_remote, fd));
-    
-    if (close(fd)) {
-	output(o_finish, _("Error writing to temporary file: %s\n"), 
-	       strerror(errno));
-    } 
-    else if (!run_editor(fname)) {
-	int upload_okay = 0;
-
-	fd = open(fname, O_RDONLY | OPEN_BINARY_FLAGS);
-	if (fd < 0) {
-	    output(o_finish, 
-		   _("Could not re-open temporary file: %s\n"),
-		   strerror(errno));
-	} else {
-	    do {
-		output(o_transfer, _("Uploading changes to `%s'"), 
-		       real_remote);
-		/* FIXME: conditional PUT using fetched Etag/modtime if
-		 * !can_lock */
-		if (out_handle(ne_put(session, real_remote, fd))) {
-		    upload_okay = 1;
-		} else {
-		    /* TODO: offer to save locally instead */
-		    printf(_("Try uploading again (y/n)? "));
-		    if (!yesno()) {
-			upload_okay = 1;
-		    }
-		}
-	    } while (!upload_okay);
-	    close(fd);
-	}
-    }
-    
-    if (unlink(fname)) {
-	printf(_("Could not delete temporary file %s:\n%s\n"), fname,
-	       strerror(errno));
-    }	       
-
-    /* UNLOCK it again whether we succeed or failed in our mission */
-    if (can_lock) {
-	output(o_start, "Unlocking `%s':", remote);
-	out_result(ne_unlock(session,lock));
-	ne_lock_remove(lock_session, lock);
-	ne_lock_free(lock);
-    }
-
-    goto edit_bail;
-edit_close:
-    close(fd);
-edit_bail:
-    free(real_remote);
-    return;
-}
-
 static void simple_put(const char *local, const char *remote)
 {
     int fd = open(local, O_RDONLY | OPEN_BINARY_FLAGS);
@@ -1186,7 +973,7 @@ static void simple_put(const char *local, const char *remote)
     }
 }
 
-void execute_put(const char *local, const char *remote)
+static void execute_put(const char *local, const char *remote)
 {
     char *real_remote;
     if (remote == NULL) {
@@ -1198,8 +985,54 @@ void execute_put(const char *local, const char *remote)
     free(real_remote);
 }
 
+/* A bit like Haskell's map function... applies func to each
+ * of the first argc items in argv */
+static void 
+map_multi(void (*func)(const char *), int argc, const char *argv[])
+{
+    int n;
+    for(n = 0; n < argc; n++) {
+	(*func)(argv[n]);
+    }
+}
+
+static void multi_copy(int argc, const char *argv[])
+{
+    do_copymove(argc, argv, _("copying"), _("copy"), simple_copy);
+}
+
+static void multi_move(int argc, const char *argv[])
+{
+    do_copymove(argc, argv, _("moving"), _("move"), simple_move);
+}
+
+static void multi_mkcol(int argc, const char *argv[])
+{
+    map_multi(execute_mkcol, argc, argv);
+}
+
+static void multi_delete(int argc, const char *argv[])
+{
+    map_multi(execute_delete, argc, argv);
+}
+
+static void multi_rmcol(int argc, const char *argv[])
+{
+    map_multi(execute_rmcol, argc, argv);
+}
+
+static void multi_less(int argc, const char *argv[])
+{
+    map_multi(execute_less, argc, argv);
+}
+
+static void multi_cat(int argc, const char *argv[])
+{
+    map_multi(execute_cat, argc, argv);
+}
+
 /* this is getting too easy */
-void multi_mput(int argc, const char *argv[])
+static void multi_mput(int argc, const char *argv[])
 {
     for(; argv[0] != NULL; argv++) {
 	char *remote;
@@ -1209,14 +1042,14 @@ void multi_mput(int argc, const char *argv[])
     }    
 }
 
-void multi_mget(int argc, const char *argv[])
+static void multi_mget(int argc, const char *argv[])
 {
     for(; argv[0] != NULL; argv++) {
 	execute_get(argv[0], NULL);
     }
 }
 
-void execute_chexec(const char *val, const char *remote)
+static void execute_chexec(const char *val, const char *remote)
 {
     static const ne_propname execprop = 
     { "http://apache.org/dav/props/", "executable" };
@@ -1256,7 +1089,7 @@ void execute_chexec(const char *val, const char *remote)
 
 }
 
-void execute_lpwd(void)
+static void execute_lpwd(void)
 {
     char pwd[BUFSIZ];
     if (getcwd(pwd, BUFSIZ) == NULL) {
@@ -1266,16 +1099,18 @@ void execute_lpwd(void)
     }
 }
 
+
 /* Using a hack, we get zero-effort multiple-argument 'lls' */
-void execute_lls(int argc, char * const argv[])
+static void execute_lls(int argc, const char **argv)
 {
+    /* nasty cast; but these describe the same array of pointers, just
+     * with different constness, so it should be okay. */
+    char *const *vpargs = (char *const *)argv;
     int pid;
     pid = fork();
     switch (pid) {
     case 0: /* child...  */
-	/* The implicit typecast munging might screw things up in
-	 * weird and wonderful ways, here? */
-	execvp("ls", &argv[-1]);
+	execvp("ls", &vpargs[-1]);
 	printf(_("Error executing ls: %s\n"), strerror(errno));
 	exit(-1);
 	break;
@@ -1289,7 +1124,7 @@ void execute_lls(int argc, char * const argv[])
     return;
 }
 
-void execute_lcd(const char *p)
+static void execute_lcd(const char *p)
 {
     const char *real_path;
     if (p) {
@@ -1307,26 +1142,26 @@ void execute_lcd(const char *p)
     }
 }
 
-void execute_pwd(void)
+static void execute_pwd(void)
 {
-    printf(_("Current collection is `%s', on server `%s'\n"), path, 
-	    server_hostname);
+    char *uri;
+    server.path = path;
+    uri = ne_uri_unparse(&server);
+    printf(_("Current collection is `%s'.\n"), uri);
+    free(uri);
 }
 
-void execute_cd(const char *newpath)
+static void execute_cd(const char *newpath)
 {
     char *real_path;
     int is_swap = false;
     if (strcmp(newpath, "-") == 0) {
+	if (!old_path) {
+	    printf(_("No previous collection.\n"));
+	    return;
+	}
 	is_swap = true;
 	real_path = old_path;
-    } else if (path == NULL) {
-	/* The first time we are executing a CD. */
-	if (1 || uri_has_trailing_slash(newpath)) {
-	    real_path = ne_strdup(newpath);
-	} else {
-	    CONCAT2(real_path, newpath, "/");
-	}
     } else {
 	real_path = resolve_path(path, newpath, true);
     }
@@ -1356,7 +1191,7 @@ static void display_help_message(void)
 	     "more=less, quit=exit=bye\n"));
 }
 
-void execute_help(const char *arg)
+static void execute_help(const char *arg)
 {
     if (!arg) {
 	display_help_message();
@@ -1373,7 +1208,7 @@ void execute_help(const char *arg)
     }
 }
 
-void execute_echo(int count, const char **args)
+static void execute_echo(int count, const char **args)
 {
     const char **pnt;
     for(pnt = args; *pnt != NULL; pnt++) {
@@ -1382,3 +1217,118 @@ void execute_echo(int count, const char **args)
     putchar('\n');
 }
 
+void execute_version(void)
+{
+    printf(PACKAGE " " VERSION "\n%s\n", ne_version_string());
+#ifdef HAVE_LIBREADLINE
+    printf("readline %s\n", rl_library_version);
+#endif
+}
+
+#if defined(__GNUC__)
+#define T0(x) {take0: x}
+#define T1(x) {take1: x}
+#define T2(x) {take2: x}
+#define T3(x) {take3: x}
+#define TV(x) {takeV: x}
+#else
+/* cast to be like first member of union. */
+#define Tn(x) { ((void (*)(void)) x) }
+#define T0(x) Tn(x)
+#define T1(x) Tn(x)
+#define T2(x) Tn(x)
+#define T3(x) Tn(x)
+#define TV(x) Tn(x)
+#endif
+
+/* C1: connected, 1-arg function C2: connected, 2-arg function
+ * U0: disconnected, 0-arg function. */
+#define C1(x,c,h) { cmd_##x, true, 1, 1, parmscope_remote, T1(execute_##x),c,h }
+#define U0(x,h) { cmd_##x, false, 0, 0, parmscope_none, T0(execute_##x),#x,h }
+#define UO1(x,c,h) { cmd_##x, false, 0, 1, parmscope_none, T1(execute_##x),c,h }
+#define C2M(x,c,h) { cmd_##x, true, 2, CMD_VARY, parmscope_remote, TV(multi_##x),c,h }
+#define C1M(x,c,h) { cmd_##x, true, 1, CMD_VARY, parmscope_remote, TV(multi_##x),c,h }
+
+/* commands[] is not static because it would mean adding a bunch of
+ * prototypes for execute_* etc, and declaring this at the top of the
+ * file. */
+
+/* Separate structures for commands and command names. */
+/* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND */
+const struct command commands[] = {
+    { cmd_ls, true, 0, 1, parmscope_remote, T1(execute_ls), 
+      N_("ls [path]"), N_("List contents of current [or other] collection") },
+    C1(cd, N_("cd path"), N_("Change to specified collection")),
+    { cmd_pwd, true, 0, 0, parmscope_none, T0(execute_pwd),
+      "pwd", N_("Display name of current collection") },
+    { cmd_put, true, 1, 2, parmscope_none, T2(execute_put),
+      N_("put local [remote]"), N_("Upload local file") },
+    { cmd_get, true, 1, 2, parmscope_none, T2(execute_get),
+      N_("get remote [local]"), N_("Download remote resource") },
+    C1M(mget, N_("mget remote..."), N_("Download many remote resources")),
+    { cmd_mput, true, 1, CMD_VARY, parmscope_local, TV(multi_mput), 
+      N_("mput local..."), N_("Upload many local files") },
+    C1(edit, N_("edit resource"), N_("Edit given resource")),
+    C1M(less, N_("less remote..."), N_("Display remote resource through pager")), 
+    C1M(mkcol, N_("mkcol remote..."), N_("Create remote collection(s)")), 
+    C1M(cat, N_("cat remote..."), N_("Display remote resource(s)")), 
+    C1M(delete, N_("delete remote..."), N_("Delete non-collection resource(s)")),
+    C1M(rmcol, N_("rmcol remote..."), N_("Delete remote collections and ALL contents")),
+    C2M(copy, N_("copy source... dest"), N_("Copy resource(s) from source to dest")), 
+    C2M(move, N_("move source... dest"), N_("Move resource(s) from source to dest")),
+    
+/* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND */
+    
+    C1(lock, N_("lock resource"), N_("Lock given resource")),
+    C1(unlock, N_("unlock resource"), N_("Unlock given resource")),
+    C1(discover, N_("discover resource"), N_("Display lock information for resource")),
+    C1(steal, N_("steal resource"), N_("Steal lock token for resource")),
+    { cmd_showlocks, true, 0, 0, parmscope_none, T0(execute_showlocks),
+      "showlocks", N_("Display list of owned locks") },
+    
+#if 0
+    C1(propedit, "propedit resource", "Enter property editor for resource"),
+#endif
+    C1(propnames, "propnames res", "Names of properties defined on resource"),
+
+    { cmd_chexec, true, 2, 2, parmscope_none, T2(execute_chexec),
+      N_("chexec [+|-] remote"), N_("Change isexecutable property of resource") },
+    
+    { cmd_propget, true, 1, 2, parmscope_none, T2(execute_propget),
+      N_("propget res [propname]"), 
+      N_("Retrieve properties of resource") },
+    { cmd_propdel, true, 2, 2, parmscope_none, T2(execute_propdel),
+      N_("propdel res propname"), 
+      N_("Delete property from resource") },
+    { cmd_propset, true, 3, 3, parmscope_none, T3(execute_propset),
+      N_("propset res propname value"),
+      N_("Set property on resource") },
+
+    { cmd_set, false, 0, 2, parmscope_none, T2(execute_set), 
+      N_("set [option] [value]"), N_("Set an option, or display options") },
+    { cmd_open, false, 1, 1, parmscope_none, T1(open_connection), 
+      "open URL", N_("Open connection to given URL") },
+    { cmd_close, true, 0, 0, parmscope_none, T0(close_connection), 
+      "close", N_("Close current connection") },
+    { cmd_echo, false, 1, CMD_VARY, parmscope_remote, TV(execute_echo), 
+      "echo", NULL },
+    { cmd_quit, false, 0, 1, parmscope_none, T1(NULL), "quit",
+      N_("Exit program") },
+    /* Unconnected operation, 1 mandatory argument */
+    { cmd_unset, false, 1, 2, parmscope_none, T2(execute_unset), 
+      N_("unset [option] [value]"), N_("Unsets or clears value from option.") },
+    /* Unconnected operation, 0 arguments */
+    UO1(lcd, N_("lcd [directory]"), N_("Change local working directory")), 
+    { cmd_lls, false, 0, CMD_VARY, parmscope_local, TV(execute_lls), 
+      N_("lls [options]"), N_("Display local directory listing") },
+    U0(lpwd, N_("Print local working directory")),
+    { cmd_logout, true, 0, 0, parmscope_none, T0(execute_logout), "logout",
+      N_("Logout of authentication session") },
+    UO1(help, N_("help [command]"), N_("Display help message")), 
+    U0(version, NULL),
+    { cmd_unknown, 0 } /* end-of-list marker, DO NOT move */
+
+/* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND. */
+
+};    
+
