@@ -139,7 +139,7 @@ static void display_ls_line(struct resource *res)
 	restype = "???"; break;
     }
     
-    if (uri_has_trailing_slash(res->uri)) {
+    if (ne_path_has_trailing_slash(res->uri)) {
 	res->uri[strlen(res->uri)-1] = '\0';
     }
     name = strrchr(res->uri, '/');
@@ -149,7 +149,7 @@ static void display_ls_line(struct resource *res)
 	name = res->uri;
     }
 
-    name = uri_unescape(name);
+    name = ne_path_unescape(name);
 
     if (res->type == resr_error) {
 	printf(_("Error: %-30s %d %s\n"), name, res->error_status,
@@ -201,25 +201,34 @@ static void results(void *userdata, const char *uri,
 {
     struct fetch_context *ctx = userdata;
     struct resource *current, *previous, *newres;
-    const char *clength, *modtime, *isexec, *abspath;
+    const char *clength, *modtime, *isexec;
     const ne_status *status = NULL;
+    ne_uri u;
 
     NE_DEBUG(NE_DBG_HTTP, "Uri: %s\n", uri);
 
     newres = ne_propset_private(set);
-    abspath = uri_abspath(uri);
-
-    NE_DEBUG(NE_DBG_HTTP, "Abspath %s in %s\n", abspath, ctx->target);
     
-    if (uri_compare(ctx->target, abspath) == 0 && !ctx->include_target) {
+    if (ne_uri_parse(uri, &u))
+	return;
+    
+    if (u.path == NULL) {
+	ne_uri_free(&u);
+	return;
+    }
+
+    NE_DEBUG(NE_DBG_HTTP, "URI path %s in %s\n", u.path, ctx->target);
+    
+    if (ne_path_compare(ctx->target, u.path) == 0 && !ctx->include_target) {
 	/* This is the target URI */
 	NE_DEBUG(NE_DBG_HTTP, "Skipping target resource.\n");
 	/* Free the private structure. */
 	free(newres);
+	ne_uri_free(&u);
 	return;
     }
 
-    newres->uri = ne_strdup(abspath);
+    newres->uri = ne_strdup(u.path);
 
     clength = ne_propset_value(set, &ls_props[0]);    
     modtime = ne_propset_value(set, &ls_props[1]);
@@ -279,6 +288,8 @@ static void results(void *userdata, const char *uri,
 	*ctx->list = newres;
     }
     newres->next = current;
+
+    ne_uri_free(&u);
 }
 
 static int end_element(void *userdata, const struct ne_xml_elm *elm, const char *cdata)
